/* 
 * ufdbguardd.c - URLfilterDB
 *
 * ufdbGuard is copyrighted (C) 2005-2013 by URLfilterDB with all rights reserved.
 * 
 * Parts of the ufdbGuard daemon are based on squidGuard.
 *
 * Multithreaded ufdbGuard daemon
 *
 * $Id: ufdbguardd.c,v 1.107 2013/01/17 17:29:46 root Exp root $
 */

#undef UFDB_FREE_MEMORY

#include "ufdb.h"
#include "sg.h"
#include "ufdbdb.h"
#include "ufdbchkport.h"
#include "httpsQueue.h"
#include "httpserver.h"

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>
#include <ctype.h>
#include <signal.h>
#include <libgen.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <netinet/tcp.h>
#include <netinet/in.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <sys/utsname.h>
#ifdef HAVE_SYS_SYSCALL_H
#include <sys/syscall.h>
#endif
#ifdef __linux__
#include <linux/unistd.h>
#endif
#include <sys/socket.h>
#if HAVE_UNIX_SOCKETS
#include <sys/un.h>
#endif
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#ifdef __linux__
#include <execinfo.h>
#endif

#ifdef _POSIX_PRIORITY_SCHEDULING
#include <sched.h>
#endif

#define FULL 1
#define EMPTY 0

pthread_mutex_t sighup_mutex = UFDB_STATIC_MUTEX_INIT;
#ifdef UFDB_HAVE_NATIVE_RWLOCK_MUTEX
pthread_rwlock_t TheDatabaseLock;
#endif

#ifdef UFDB_HAVE_NATIVE_RWLOCK_MUTEX
extern pthread_rwlock_t  TheDynamicSourcesLock;
#else
extern pthread_mutex_t   TheDynamicSourcesLock;
#endif

#define UFDB_MAX_WORKERS	UFDB_MAX_THREADS
static int     n_workers = 65;

static int     my_argc;
static char ** my_argv;
static char ** my_envp;
static char *  configFile = NULL;
static struct timeval  start_time;

static int       badSignal = 0;
static pthread_t badSignalThread;

char * globalPidFile = UFDBGUARDD_PID_FILE;

extern int globalPid;

extern int globalDebugTimeDelta;
extern int sig_other;
extern char ** globalArgv;
extern char ** globalEnvp;

extern int httpsConnectionCacheSize;

static int        portNumCmdLine = -1;

static int        testConfig = 0;
static int        runAsDaemon = 1;
static int        testMode = 0;
static int        parseOnly = 0;
static pthread_t  threadedSignalHandler = (pthread_t) 837;
static pthread_t  sockethandler = (pthread_t) 837;
static pthread_t  dyn_userlist_handler = (pthread_t) 837;
static pthread_t  workers[UFDB_MAX_WORKERS] = { (pthread_t) 837, (pthread_t) 837 };
static pthread_t  httpsthr[UFDB_NUM_HTTPS_VERIFIERS] = { (pthread_t) 837, (pthread_t) 837 };
static pid_t      httpDaemonPid = 0;
static int        allowedToRemovePidFile = 0;

volatile unsigned long UFDBlookupCounter = 0;
volatile unsigned long UFDBlookupHttpsCounter = 0;
volatile unsigned long UFDBblockCounter = 0;
volatile unsigned long UFDBtestBlockCounter = 0;
volatile unsigned long UFDBuncategorisedCounter = 0;
volatile unsigned long UFDBsafesearchCounter = 0;
volatile unsigned long UFDBedufilterCounter = 0;
volatile unsigned long UFDBuploadSeqNo = 0;

static void uploadStatistics( char * reason );
static void * worker_main( void * ptr );
static void * dynamic_userlist_updater_main( void * ptr );

#ifdef UFDB_DEBUG
extern int yydebug;
#endif


static void usage( char option )
{
   if (option != '\0')
      fprintf( stderr, "unknown option '-%c'\n", option );

   fprintf( stderr, "usage: ufdbguardd [-A] [-d] [-v] [-r] [-R] [-T] [-h] [-w 32-129] [-p port] [-c <configfile>]\n" );
   fprintf( stderr, "-v      print version\n" );
   fprintf( stderr, "-c FN   use configuration file FN\n" );
   fprintf( stderr, "-T      test mode; do not block, only log which URLs might be blocked\n" );
   fprintf( stderr, "-w N    use N worker threads (default: %d)\n", n_workers );
   fprintf( stderr, "-p N    use TCP port N to listen on\n" );
   fprintf( stderr, "-U user run with privileges of \"user\"\n" );
   fprintf( stderr, "-d      extra debug information in log file\n" );
   fprintf( stderr, "-R      debug RE matching\n" );
   fprintf( stderr, "-r      log all redirections\n" );
   fprintf( stderr, "-N      do not analyse uncategorised URLs (not recommended)\n" );
   fprintf( stderr, "-h      print this help text\n" );
   fprintf( stderr, "-C verify  verify a configuration file; must be the last option\n" );

   if (option != '\0')
      exit( 1 );
   else
      exit( 0 );
}


static void writePidFile( void )
{
   FILE * fp;

   (void) unlink( globalPidFile );
   fp = fopen( globalPidFile, "w" );
   if (fp == NULL)
      ufdbLogError( "cannot write to PID file %s", globalPidFile );
   else
   {
      fprintf( fp, "%d\n", globalPid );
      fclose( fp );
      allowedToRemovePidFile = 1;
   }
}


static void removePidFile( void )
{
   FILE * fp;
#if HAVE_UNIX_SOCKETS
   char   unixsocket[64];
#endif

   if (UFDBglobalDebug)
      ufdbLogMessage( "removing pid file  allowed=%d", allowedToRemovePidFile );

   /* When the initialisation of the socket failed there is most likely
    * already an ufdbguardd daemon running and we may not remove
    * the pid file nor kill any ufdbhttpd daemon.
    */
   if (!allowedToRemovePidFile)
      return;

#if HAVE_UNIX_SOCKETS
   sprintf( unixsocket, "/tmp/ufdbguardd-%05d", UFDBglobalPortNum );
   if (unlink( unixsocket ) < 0  &&  errno != ENOENT)
      ufdbLogError( "cannot remove socket %s: %s", unixsocket, strerror(errno) );
#endif

   (void) unlink( globalPidFile );

   allowedToRemovePidFile = 0;

   if (UFDBglobalHttpdPort > 0)
   {
      UFDBraisePrivileges( UFDBglobalUserName, "kill ufdbhttpd" );
      fp = fopen( UFDBHTTPD_PID_FILE, "r" );
      if (fp != NULL)
      {
         if (fscanf( fp, "%d", &httpDaemonPid ) == 1)
	 {
	    ufdbLogMessage( "sending TERM signal to the HTTP daemon (pid=%d)", httpDaemonPid );
	    if (kill( httpDaemonPid, SIGTERM ) < 0)
	       ufdbLogError( "cannot kill pid %d: %s", httpDaemonPid, strerror(errno) );
	    /* TODO: find out why ufdbhttpd does not act on the TERM signal... */
	    usleep( 210000 );
	    (void) kill( httpDaemonPid, SIGKILL );
	 }
         fclose( fp );
      }
      UFDBdropPrivileges( UFDBglobalUserName );
   }
}


static void startHttpDaemon( void )
{
   int    i;
   char * argv[24];
   struct stat statbuf;
   char   portStr[16];
   char   httpdbin[1024];

   sprintf( httpdbin, "%s/%s", DEFAULT_BINDIR, "ufdbhttpd" );
   if (stat( httpdbin, &statbuf ) < 0)
   {
      ufdbLogError( "cannot find the HTTP daemon executable (%s): %s  *****", httpdbin, strerror(errno) );
      return;
   }

   /* go back to user root to allow ufdbhttpd to use privileged ports */
   UFDBraisePrivileges( UFDBglobalUserName, "start ufdbhttpd" );

   httpDaemonPid = fork();
   if (httpDaemonPid == 0)	/* child */
   {
      UFDBcloseFilesNoLog();
      i = 0;
      argv[i++] = httpdbin;
      if (UFDBglobalDebug || UFDBglobalDebugHttpd)
         argv[i++] = "-d";
      argv[i++] = "-p";
      sprintf( portStr, "%d", UFDBglobalHttpdPort );
      argv[i++] = portStr;
      argv[i++] = "-l";
      if (UFDBglobalLogDir == NULL)
	 argv[i++] = DEFAULT_LOGDIR;
      else
	 argv[i++] = UFDBglobalLogDir;
      argv[i++] = "-I";
      if (UFDBglobalHttpdImagesDirectory[0] == '\0')
	 strcpy( UFDBglobalHttpdImagesDirectory, DEFAULT_IMAGESDIR );
      argv[i++] = UFDBglobalHttpdImagesDirectory;
      argv[i++] = "-i";
      argv[i++] = UFDBglobalHttpdInterface;
      if (UFDBglobalUserName[0] != '\0')
      {
	 argv[i++] = "-U";
	 argv[i++] = UFDBglobalUserName;
      }
      argv[i] = NULL;
      ufdbLogMessage( "starting HTTP daemon ..." );
      execv( httpdbin, argv );
      ufdbLogError( "failed starting http daemon: execv failed: %s", strerror(errno) );
      _exit( 2 );
   }

   UFDBdropPrivileges( UFDBglobalUserName );

   if (httpDaemonPid < 0)
   {
      ufdbLogError( "cannot start HTTP daemon: fork failed: %s", strerror(errno) );
      httpDaemonPid = 0;
   }
   else
   {
#if 0
      int        status;
      pid_t      pid;
#endif

      ufdbLogMessage( "ufdbhttpd is started.  port=%d interface=%s images=%s",
                      UFDBglobalHttpdPort, UFDBglobalHttpdInterface, UFDBglobalHttpdImagesDirectory );
#if 0
      usleep( 300000 );
      while ((pid = waitpid( -1, &status, WNOHANG )) > 0)
	 ;
#endif
   }
}


static void adjustNumberOfThreads( void )
{
   int              i;
   pthread_attr_t   attr;

   /* IF the check_https_tunnel_option is aggressive we need more worker threads */
   /* Note that UFDB_MAX_THREADS is 129 so with the -w option we can have 129 worker threads */
   if (UFDBgetTunnelCheckMethod() == UFDB_API_HTTPS_CHECK_AGGRESSIVE  &&  n_workers < 129)
   {
      /* create more worker threads */
      i = n_workers;
      n_workers = 129;
      pthread_attr_init( &attr );
      pthread_attr_setscope( &attr, PTHREAD_SCOPE_SYSTEM );
      pthread_attr_setstacksize( &attr, 248 * 1024 );
#if HAVE_PTHREAD_ATTR_SETGUARDSIZE && UFDB_DEBUG
      pthread_attr_setguardsize( &attr, 8 * 1024 );
#endif
      for (; i < n_workers; i++)
      {
	 pthread_create( &workers[i], &attr, worker_main, (void *) ((long) i) );
#ifdef _POSIX_PRIORITY_SCHEDULING
	 sched_yield();
#else
	 usleep( 10000 );
#endif
      }
      ufdbLogMessage( "#threads is increased to %d because check-proxy-tunnels is set to \"aggressive\"", n_workers );
   }
}


static void Usr1SignalCaught( int sig )
{
   FILE * fp;
   int    oldStatus;

   ufdbLogMessage( "USR1 signal received for logfile rotation" );
   oldStatus = UFDBchangeStatus( UFDB_API_STATUS_ROLLING_LOGFILE );
   UFDBrotateLogfile();
   if (UFDBglobalHttpdPort > 0)
   {
      fp = fopen( UFDBHTTPD_PID_FILE, "r" );
      if (fp != NULL)
      {
	 if (1 != fscanf( fp, "%d", &httpDaemonPid ))
	    httpDaemonPid = -1;
	 fclose( fp );
      }
      else
	 httpDaemonPid = -1;

      if (httpDaemonPid > 0)
      {
	 int retval;

	 ufdbLogMessage( "sending USR1 signal to ufdbhttpd (pid=%d)", httpDaemonPid );
	 /* TODO: use setreuid() */
	 retval = kill( httpDaemonPid, SIGUSR1 );
	 if (retval != 0)
	    ufdbLogError( "cannot send USR1 signal to httpd daemon: %s", strerror( errno ) );
      }
   }
   ufdbLogMessage( "USR1 signal received for logfile rotation" );
   UFDBchangeStatus( oldStatus );
}


static void Usr2SignalCaught( int sig )
{
   int    oldStatus;

   ufdbLogMessage( "USR2 signal received to trigger monitoring commands" );
   oldStatus = UFDBchangeStatus( UFDB_API_STATUS_UPDATE );
   UFDBchangeStatus( oldStatus );
   ufdbLogMessage( "status update done" );
}


/*
 * OpenBSD is very picky with its implementation of signals and
 * does not allow violation of the POSIX standard that says that
 * a signal handler may NOT raise a signal, so calling pthread_kill
 * is not allowed.  
 * The signal handlers below do nothing.  They are installed to
 * tell the OS to send the signals to the process and the special
 * signal interception threads will handle the signals asynchroneously.
 *
 * These signal handlers can do nothing, not even call 
 * ufdbLogMessage because it uses a mutex.
 */
static void catchHUPSignal( int signal )
{
}


static void catchUSR1Signal( int signal )
{
}


static void catchUSR2Signal( int signal )
{
}


static void catchChildSignal( int signal )
{
   /* For some reason, on OpenBSD the SIGCHLD does not get through
    * to the signal_handler_thread() so we must do a waitpid() here.
    */
#if defined(__OpenBSD__) || defined(__NetBSD__)  || defined(__FreeBSD__)
   int        status;
   pid_t      pid;

   while ((pid = waitpid( -1, &status, WNOHANG )) > 0)
      ;
#endif
}


void catchAbortSignal( int signum )
{
   UFDBglobalReconfig = 1;
   UFDBglobalTerminating = 1;
   (void) alarm( 0 );

   ufdbLogMessage( "ABORT signal was sent. exiting..." );

   removePidFile();
   UFDBchangeStatus( UFDB_API_STATUS_TERMINATED );
   exit( 3 );
}


void catchTermSignal( int signum )
{
   /* do nothing. sigwait() in a thread will deal with the signal */
}


static void TermSignalCaught( int signum )
{
   UFDBglobalTerminating = 1;
   UFDBglobalReconfig = 1;
   (void) alarm( 0 );

   ufdbLogMessage( "received %s signal", signum == SIGINT ? "INT" : "TERM" );
   uploadStatistics( signum == SIGINT ? "INT" : "TERM" );

   UFDBchangeStatus( UFDB_API_STATUS_TERMINATED );

   usleep( 100000 );
   pthread_cancel( sockethandler );

   /* we are inside the signal_handler_thread and prefer that the main thread does all the cleanup */
   /* do 8 short sleeps instead of one large sleep since a few of them will be interrupted by signals */
   sleep( 1 );		
   sleep( 1 );	
   sleep( 1 );	
   sleep( 1 );	
   sleep( 1 );	
   sleep( 1 );	
   sleep( 1 );	
   sleep( 1 );	

   /* close all connections with clients */
   UFDBcloseFilesNoLog();

   removePidFile();
   ufdbLogMessage( "TermSignalCaught: bye bye" );
   _exit( 0 );
}


static char * my_fast_fgets_nonl( char * s, int size, FILE * stream )
{
   char * buf;
   int    ch;

   buf = s;
   while ((ch = getc_unlocked(stream)) != EOF  &&  --size > 0)
   {
      if (ch == '\n'  ||  ch == '\r')
         break;
      *buf++ = ch;
   }
   *buf = '\0';
   if (ch == EOF  &&  buf == s)
      return NULL;
   return s;
}


void interruptPstack( int dummy )
{
   ufdbLogError( "waited too long for output of pstack.... aborting now!" );
   UFDBchangeStatus( UFDB_API_STATUS_FATAL_ERROR );
   _exit( 5 );
}


void BadSignalCaught( int signum )	
{
#ifdef __linux__
   int       i;
   int       num;
   int       i_am_thread;
   char      me[32];
#endif
   
#ifdef __linux__
   /* find out which thread has a signal */
   i_am_thread = 0;
   sprintf( me, "pid %d", (int) getpid() );
   if (badSignalThread == threadedSignalHandler)
   {
      strcpy( me, "thread signal-handler" );
      i_am_thread = 1;
   }
   else if (badSignalThread == sockethandler)
   {
      strcpy( me, "thread socket-handler" );
      i_am_thread = 1;
   }
   else if (badSignalThread == dyn_userlist_handler)
   {
      strcpy( me, "thread dynamic-userlist-handler" );
      i_am_thread = 1;
   }
   else 
   {
      num = UFDB_NUM_HTTPS_VERIFIERS;
      for (i = 0; i < num; i++)
         if (badSignalThread == httpsthr[i])
	 {
	    sprintf( me, "thread https-verifier-%02d", i );
	    i_am_thread = 1;
	    break;
	 }
      num = n_workers;
      for (i = 0; i < num; i++)
         if (badSignalThread == workers[i])
	 {
	    sprintf( me, "thread worker-%02d", i );
	    i_am_thread = 1;
	    break;
	 }
   }
   ufdbLogMessage( "%s caught signal %d.", me, signum );

#if 1
   /* let workers finish their jobs */
   usleep( 500000 );
#endif

   if (strcmp( UFDB_GDB_PATH, "no" ) == 0)
   {
      ufdbLogMessage( "ufdb-pstack is not called since the gdb package was not installed" );
   }
   else
   {
      FILE * fp;
      char   cmd[512];
      char   buffer[2048];

      ufdbSetSignalHandler( SIGALRM, interruptPstack );
      alarm( 6 );

      sprintf( cmd, DEFAULT_BINDIR "/ufdb-pstack %d 2>&1", getpid() );

      /* gdb sometimes needs root permissions, so the popen() is done as root */
      UFDBraisePrivileges( UFDBglobalUserName, "open pipe for ufdb-pstack" );
      errno = 0;
      fp = popen( cmd, "r" );
      if (fp == NULL)
         ufdbLogError( "could not open pipe and execute \"%s\": %s", cmd, strerror(errno) );

      while (my_fast_fgets_nonl( buffer, 2040, fp ) != NULL)
      {
         ufdbLogMessage( "pstack  %s", buffer );
      }
      pclose( fp );
      ufdbLogMessage( "pstack  END" );
      UFDBdropPrivileges( UFDBglobalUserName );
   }
#else
   ufdbLogMessage( "BadSignalCaught: one of the threads caused signal %d", signum );
#endif

   UFDBchangeStatus( UFDB_API_STATUS_FATAL_ERROR );
   /* 
    * ufdbguardd is a multithreaded application and other threads may also 
    * call UFDBchangeStatus() ...  We have seen the status being changed 
    * to ERROR and immediately followed by another status change to RELOADED.
    * To try to make it more robust we do a short sleep and set the status again.
    */
   usleep( 500000 );
   UFDBchangeStatus( UFDB_API_STATUS_FATAL_ERROR );

#if 0
   if (UFDBglobalDebug)
      ufdbLogMessage( "BadSignalCaught: going to kill socket-handler..." );
   pthread_cancel( sockethandler );

   /* We prefer that the main thread does all the cleanup. */
   /* So we do 8 short sleeps instead of one large sleep since a few of them will be interrupted by signals. */
   sleep( 2 );		
   sleep( 2 );	
   sleep( 2 );	
   sleep( 2 );	
   sleep( 2 );	
   sleep( 2 );	
   sleep( 2 );	
   sleep( 2 );	
   /* all threads including myself should be dead now */

   ufdbLogMessage( "BadSignalCaught: ufdbguardd aborts" );
   UFDBchangeStatus( UFDB_API_STATUS_TERMINATED );

   removePidFile();
   kill( globalPid, SIGABRT );

   sleep( 2 );
   exit( 2 );
#endif
}


void ufdbCatchBadSignal( int signum )	
{
   /* Note that we are inside a signal handler and can do almost nothing. */

   if (badSignal != signum)
   {
      UFDBglobalReconfig = 1;
      UFDBglobalTerminating = 1;
      (void) alarm( 0 );

      ufdbLogMessage( "ufdbCatchBadSignal: received signal %d", signum );

      badSignal = signum;
      badSignalThread = pthread_self();

      /* Cancel the sockethandler thread so that the main thread will wake up and 
       * wind up.
       */
      pthread_cancel( sockethandler );

      /* If for any reason the termination by main fails, we have to call exit() ourselves. */
      sleep( 2 );
      sleep( 2 );
      sleep( 2 );
      sleep( 2 );
      sleep( 2 );
      sleep( 2 );
      sleep( 2 );
      sleep( 2 );
      sleep( 2 );
      sleep( 2 );
      sleep( 2 );
      ufdbLogError( "ufdbCatchBadSignal: calling _exit() after waiting for 20 seconds for main to terminate ufdbguardd" );
      exit( 1 );
   }
}


static void HupSignalCaught( int sig )
{
   time_t          t;
   struct tm       thetime;
   int             status;
   pid_t           pid;
   FILE *          fp;
#ifdef UFDB_HAVE_NATIVE_RWLOCK_MUTEX
   int             ret;
   struct timespec DatabaseLockTimeout;
#else
   unsigned int    sleep_period;
#endif

   if (pthread_mutex_trylock( &sighup_mutex ) != 0)
   {
      ufdbLogMessage( "HUP signal is already being processed" );
      return;
   }

   /* going to reload the configuration but first wait for others to finish */
   while (UFDBglobalReconfig)
      usleep( 100000 );

   UFDBglobalReconfig = 1;
   alarm( 0 );

   ufdbLogMessage( "HUP signal received to reload the configuration and database" );
   UFDBchangeStatus( UFDB_API_STATUS_RELOADING );
   uploadStatistics( "HUP" );

   ufdbLogMessage( "statistics: %lu URL lookups (%lu https). %lu URLs blocked. "
		   "%lu tunnels detected. %lu safe searches. %lu Youtube edufilter. %lu uncategorised URLs. %lu clients.",
		   UFDBlookupCounter, 
		   UFDBlookupHttpsCounter,
		   UFDBblockCounter + UFDBtestBlockCounter,
		   UFDBglobalTunnelCounter,
		   UFDBsafesearchCounter,
		   UFDBedufilterCounter,
		   UFDBuncategorisedCounter,
		   UFDBgetNumberOfRegisteredIPs() );

   /* To prevent counting 'old' clients: 
    * if today is Wednesday, we reset the IP counters.
    */
   t = time( NULL );
   localtime_r( &t, &thetime );
   if (thetime.tm_wday == 3)
   {
      UFDBinitializeIPcounters();
   }

   /* Get the pid of the HTTP daemon. It is used later to kill it. */
   httpDaemonPid = -1;
   if (UFDBglobalHttpdPort > 0)
   {
      fp = fopen( UFDBHTTPD_PID_FILE, "r" );
      if (fp != NULL)
      {
	 if (1 != fscanf( fp, "%d", &httpDaemonPid ))
	 {
	    ufdbLogError( "file %s does not have a process id", UFDBHTTPD_PID_FILE );
	    httpDaemonPid = -1;
	 }
	 fclose( fp );
      }
      else
      {
	 ufdbLogError( "cannot open file %s: %s", UFDBHTTPD_PID_FILE, strerror(errno) );
      }
   }

   sleep( 1 );

#ifdef UFDB_HAVE_NATIVE_RWLOCK_MUTEX
   /* 
    * Synchronise with all worker threads.
    * UFDBglobalReconfig is set for the last 2 seconds so the worker threads have 
    * most likely no read-locks any more.  Try to get a write lock within 0.1 second
    * and if this fails, we will sleep 3 seconds more and acquire the rwlock with
    * pthread_rwlock_wrlock().
    *
    * We use a 0.2-second timeout since we do not want to block readers (workers)
    * too long because the flow of messages back to Squid may not be interrupted too long.
    * Note: 1,000,000,000 nanoseconds make 1 second.
    */
   ufdbLogMessage( "requesting a write lock for the in-memory database (fast mode)..." );

   DatabaseLockTimeout.tv_sec = 0;
   DatabaseLockTimeout.tv_nsec = 200000000;
   ret = pthread_rwlock_timedwrlock( &TheDatabaseLock, &DatabaseLockTimeout );
   if (ret == 0)
      ufdbLogMessage( "lock acquired to reload the URL database" );
   else
   {
      if (ret == ETIMEDOUT)
      {
	 ufdbLogMessage( "Could not obtain a lock quickly to refresh the URL database\n"
	                 "sleeping 3 seconds before attemting to acquire the lock again" );
	 sleep( 3 );
      }
      else 
	 ufdbLogError( "pthread_rwlock_timedwrlock failed with code %d", ret );

      ufdbLogMessage( "requesting a write lock for the in-memory database (slow mode)..." );

      /* 
       * Wait many seconds for a write lock ...
       * THIS *MAY* DISTURB USERS since Squid may not be getting answers back from some workers.
       */
      DatabaseLockTimeout.tv_sec = 15;
      DatabaseLockTimeout.tv_nsec = 200000000;
      ret = pthread_rwlock_timedwrlock( &TheDatabaseLock, &DatabaseLockTimeout );
      if (ret == 0)
	 ufdbLogMessage( "lock acquired to reload the URL database" );
      else
      {
	 if (ret == ETIMEDOUT)
	    ufdbLogError( "Could not acquire rw-lock. The pthread implementation is poor for acquiring write locks *****" );
	 else
	    ufdbLogError( "pthread_rwlock_wrlock failed with code %d", ret );
      }
   }

   /* We must unlock the lock ASAP since the worker threads must continue handling requests.
    * The workers are now synchronised and all know that UFDBglobalReconfig is set.
    */
   if (ret == 0)
   {
      ufdbLogMessage( "releasing the lock so that the synchronised worker threads can continue" );
      (void) pthread_rwlock_unlock( &TheDatabaseLock );
   }

#else
   /* sleep() is a very crude way to allow worker threads to finish what
    * they were doing, before we remove the database from memory,
    * but it allows us not to use mutexes for a many-reader-one-writer problem.
    */
   sleep_period = n_workers / 2;
   if (sleep_period < 12)
      sleep_period = 12;
   if (sleep_period > 20)
      sleep_period = 20;
   sleep_period -= 3+1+1;		/* we already slept a few seconds */
   ufdbLogMessage( "sleeping %d seconds to allow worker threads to finish current work", 
		   sleep_period );
   sleep( sleep_period );		
#endif

   /* TODO: reload CA certificates and use TheDatabaseLock */
   /* TODO: clear security cache */

   ufdbLogMessage( "releasing database from memory ..." );
   ufdbFreeAllMemory();
   ufdbLogMessage( "reading configuration file and database ..." );
   UFDBglobalFatalError = 0;
   UFDBglobalHttpdPort = 0;
   sgReadConfig( configFile );
   if (UFDBglobalFatalError)
   {
      ufdbLogMessage( "A FATAL ERROR OCCURRED: all requests are answered with \"OK\" (see previous lines for more details)  *****" );
   }

   /* TODO: only stop/start ufdbhttpd if the parameters have changed */
   if (httpDaemonPid > 0)
   {
      int retval;

      /* TODO: call setreuid() */
      ufdbLogMessage( "killing old ufdbhttpd (pid=%d)", httpDaemonPid );
      retval = kill( httpDaemonPid, SIGTERM );
      if (retval != 0)
	 ufdbLogError( "cannot send TERM signal to the HTTP daemon (pid=%d): %s", 
		       httpDaemonPid, strerror( errno ) );

      usleep( 400000 );
      /* be sure that the daemon is killed. */
      if (kill( httpDaemonPid, SIGKILL ) == 0)
	 usleep( 100000 );

      while ((pid = waitpid( -1, &status, WNOHANG )) > 0)
	 ;

      if (kill( httpDaemonPid, 0 ) < 0)
	 ufdbLogMessage( "HTTP daemon (pid=%d) terminated to be restarted (OK)", httpDaemonPid );
      else
	 ufdbLogError( "HTTP daemon did not terminate: %s", strerror(errno) );
   }

   if (UFDBglobalHttpdPort > 0)
      startHttpDaemon();

   UFDBlogConfig();

   adjustNumberOfThreads();

   UFDBlookupCounter = 0;
   UFDBlookupHttpsCounter = 0;
   UFDBblockCounter = 0;
   UFDBtestBlockCounter = 0;
   UFDBuncategorisedCounter = 0;
   UFDBglobalTunnelCounter = 0;
   UFDBsafesearchCounter = 0;
   UFDBedufilterCounter = 0;
   UFDBuploadSeqNo++;

   if (UFDBglobalFatalError)
   {
      UFDBchangeStatus( UFDB_API_STATUS_FATAL_ERROR );
   }
   else
   {
      UFDBchangeStatus( UFDB_API_STATUS_RELOAD_OK );
   }

#if 0
   /* delay resetting UFDBglobalReconfig to let the workers finish their loops */
   usleep( 100000 );
#endif 

   ufdbHandleAlarmForTimeEvents( UFDB_PARAM_INIT );
   if (UFDBglobalFatalError)
      ufdbLogMessage( "A FATAL ERROR OCCURRED: all requests are answered with \"OK\" (see previous lines for more details)  *****" );
   else
      ufdbLogMessage( "the new configuration and database are loaded for ufdbguardd " VERSION );

   UFDBglobalReconfig = 0;

   pthread_mutex_unlock( &sighup_mutex );

   pthread_kill( dyn_userlist_handler, SIGHUP );
}


/* 
 * Communication between the worker and other threads is done via
 * RW-locks and data is passed via a queue.
 *
 * Since this queue is a FIFO we could use a tail queue. But we don't
 * like the malloc() overhead, so therefore we use a static array.
 */

#define UFDB_MAX_FD_QUEUE   UFDB_MAX_THREADS


static struct {
   int    fd;
} q[UFDB_MAX_FD_QUEUE];

volatile static int    ihead = 0;                    /* array index for the head */
volatile static int    itail = 0;                    /* array index for the tail */
volatile static int    n_queued = 0;
static pthread_mutex_t fdq_lock = UFDB_STATIC_MUTEX_INIT;
static pthread_cond_t  empty = PTHREAD_COND_INITIALIZER;


/*
 * enqueue a new fd.
 */
static void putFdQueue( int fd )
{
   int ret;

   ret = pthread_mutex_lock( &fdq_lock );
#ifdef UFDB_DEBUG
   if (ret != 0)
      ufdbLogError( "putFdQueue: mutex_lock failed with code %d *****", ret );
#endif

   if (n_queued < UFDB_MAX_FD_QUEUE)
   {
      /*
       * insert it at the tail
       */
      q[itail].fd = fd;

      itail = (itail + 1) % UFDB_MAX_FD_QUEUE;
      n_queued++;

      /*
       * leave the critical section
       */
      ret = pthread_mutex_unlock( &fdq_lock );
#ifdef UFDB_DEBUG
      if (ret != 0)
	 ufdbLogError( "putFdQueue: mutex_unlock failed with code %d *****", ret );
#endif

      /*
       * signal anyone who is waiting
       */
      pthread_cond_broadcast( &empty );
   } 
   else    
   {
      ufdbLogFatalError( "FATAL ERROR: connection queue is full.  *****\n"
                         "There are too many ufdbgclient processes !!\n"
                         "the maximum is %d when the -w option is used\n"
      			 "closing fd %d", 
			 UFDB_MAX_FD_QUEUE, fd );
      /* there is no other option than to close the connection */
      /* TO-DO: maybe: close all queued fd's ? */
      close( fd );
   }
}


/*
 * get a fd where there is new content.
 */
static void getFdQueue( int * fd )
{
   int ret;

allover:
   ret = pthread_mutex_lock( &fdq_lock );
#ifdef UFDB_DEBUG
   if (ret != 0)
      ufdbLogError( "getFdQueue: mutex_lock failed with code %d *****", ret );
#endif

   while (1)
   {
      /*
       * if there are jobs in the queue
       */
      if (n_queued > 0)
      {
         n_queued--;
         /*
          * get the one at the head
          */
         *fd = q[ihead].fd;
         ihead = (ihead + 1) % UFDB_MAX_FD_QUEUE;

         ret = pthread_mutex_unlock( &fdq_lock );
#ifdef UFDB_DEBUG
	 if (ret != 0)
	    ufdbLogError( "getFdQueue: mutex_unlock failed with code %d *****", ret );
#endif

	 return;
      } 
      else   /* otherwise wait until there are fds available */ 
      {
         pthread_cond_wait( &empty, &fdq_lock );
         /*
          * when I'm here, I've been signaled because there
          * are jobs in the queue.  Go try and get one.
          */
	 ret = pthread_mutex_unlock( &fdq_lock );
#ifdef UFDB_DEBUG
	 if (ret != 0)
	    ufdbLogError( "getFdQueue: mutex_unlock failed with code %d *****", ret );
#endif
	 usleep( ((unsigned long) pthread_self()) % 821 );
	 goto allover;
      }
   }
}


static void daemon_accept_connections( 
   int            s, 
   int            protocol )
{
   int            n;
   int            newfd;
   fd_set         fds;
   struct timeval tv;

   /*
    * Allow that this thread can be cancelled without delays at any time.
    */
   pthread_setcancelstate( PTHREAD_CANCEL_ENABLE, NULL );
   pthread_setcanceltype( PTHREAD_CANCEL_ASYNCHRONOUS, NULL );

   while (!UFDBglobalTerminating)
   {
      FD_ZERO( &fds );
      FD_SET( s, &fds );
      errno = 0;
      n = select( s+1, &fds, (fd_set *) NULL, (fd_set *) NULL, (struct timeval *) NULL );

      if (UFDBglobalTerminating)
	 break;

      if (n > 0) 
      {
	 newfd = accept( s, NULL, NULL );
	 if (newfd < 0)
	 {
	    if (errno == EINTR)
	    {
	       continue;
	    }
	    ufdbLogError( "SRV: accept on socket failed: %s", strerror(errno) );
	    return;
	 }

	 if (protocol == AF_INET)
	 {
	    int sock_parm;

	    sock_parm = 1;
	    setsockopt( newfd, IPPROTO_TCP, TCP_NODELAY, (void *) &sock_parm, sizeof(sock_parm) );
	    sock_parm = 1;
	    setsockopt( newfd, SOL_SOCKET, SO_KEEPALIVE, (void *) &sock_parm, sizeof(sock_parm) );
	    sock_parm = 16384;
	    setsockopt( newfd, SOL_SOCKET, SO_SNDBUF, (void *) &sock_parm, sizeof(sock_parm) );
	    sock_parm = 16384;
	    setsockopt( newfd, SOL_SOCKET, SO_RCVBUF, (void *) &sock_parm, sizeof(sock_parm) );
	 }

	 /*
	  *  In case of troubles, the timeout prevents that a thread will block for ever.
	  */
#if 0
	 tv.tv_sec = 16 * 60;
	 tv.tv_usec = 0;
	 setsockopt( newfd, SOL_SOCKET, SO_RCVTIMEO, (void *) &tv, sizeof(tv) );
#endif
	 tv.tv_sec = 5;			/* 5 second send timeout */
	 tv.tv_usec = 0;
	 setsockopt( newfd, SOL_SOCKET, SO_SNDTIMEO, (void *) &tv, sizeof(tv) );

	 if (UFDBglobalDebug > 1)
	    ufdbLogMessage( "SRV: accept new fd %2d", newfd );

	 putFdQueue( newfd );
      }
   }
}


void * socket_handler_thread( void * ptr )
{
   int                 protocol;
   sigset_t            sigset;
   int                 i;
   int                 s;
   int                 sock_parm;
   int                 retval;
   long                num_cpus;
   struct utsname      sysinfo;
   pthread_attr_t      attr;
   struct sockaddr_in  addr_in;
#if HAVE_UNIX_SOCKETS
   struct sockaddr_un  addr_un;
#endif

   UFDBglobalReconfig = 1;

   /* Most signals must be blocked.
    * This is a requirement to use sigwait() in a thread.
    */
   sigemptyset( &sigset );
   sigaddset( &sigset, SIGHUP );
   sigaddset( &sigset, SIGUSR1 );
   sigaddset( &sigset, SIGUSR2 );
   sigaddset( &sigset, SIGCHLD );
   sigaddset( &sigset, SIGTERM );
   sigaddset( &sigset, SIGINT );
   sigaddset( &sigset, SIGALRM );
#if 0
   sigaddset( &sigset, SIGSEGV );	/* blocking SEGV with pthread_sigmask has undefined results */
   sigaddset( &sigset, SIGILL );
   sigaddset( &sigset, SIGBUS );
#endif
   pthread_sigmask( SIG_BLOCK, &sigset, NULL );

   gettimeofday( &start_time, NULL );

   globalArgv = my_argv;
   globalEnvp = my_envp;

   ufdbSetGlobalErrorLogFile();
   ufdbResetUnknownURLs();
   UFDBglobalFatalError = 0;
   sgReadConfig( configFile );
   /* ufdbSetGlobalErrorLogFile(); */
   UFDBlogConfig();

   adjustNumberOfThreads();

   UFDBinitHTTPSchecker();
#ifdef _POSIX_PRIORITY_SCHEDULING
   sched_yield();
#else
   usleep( 50000 );
#endif

   pthread_attr_init( &attr );
   pthread_attr_setscope( &attr, PTHREAD_SCOPE_SYSTEM );
   pthread_attr_setstacksize( &attr, 248 * 1024 );
#if HAVE_PTHREAD_ATTR_SETGUARDSIZE && UFDB_DEBUG
   pthread_attr_setguardsize( &attr, 8 * 1024 );
#endif
   for (i = 0; i < UFDB_NUM_HTTPS_VERIFIERS; i++)
   {
      pthread_create( &httpsthr[i], &attr, UFDBhttpsTunnelVerifier, (void *) ((long) i) );
#ifdef _POSIX_PRIORITY_SCHEDULING
      sched_yield();
#else
      usleep( 10000 );
#endif
   }
   usleep( 50000 );

   if (testMode)
      ufdbLogMessage( "-T option is used.  In test mode no URLs are ever blocked." );

   if (UFDBglobalFatalError)
   {
      ufdbLogError( "A FATAL ERROR OCCURRED: ALL REQUESTS ARE ANSWERED WITH \"OK\" (see previous lines for more information)  *****" );
      UFDBchangeStatus( UFDB_API_STATUS_FATAL_ERROR );
   }
   else
   {
      ufdbHandleAlarmForTimeEvents( UFDB_PARAM_INIT );
      UFDBchangeStatus( UFDB_API_STATUS_STARTED_OK );
   }

   /*
    * Create the daemon socket that the it accepts connections on.
    * if available, first try a UNIX socket and then an IP socket.
    */
#if HAVE_UNIX_SOCKETS
   protocol = AF_UNIX;
   s = socket( protocol, SOCK_STREAM, 0 );
   if (s < 0)
   {
      protocol = AF_INET;
      s = socket( protocol, SOCK_STREAM, 0 );
   }
#else
   protocol = AF_INET;
   s = socket( protocol, SOCK_STREAM, 0 );
#endif

   if (s < 0)
   {
      ufdbLogError( "cannot create daemon socket: %s", strerror(errno) );
      pthread_exit( (void *) 0 ); 
   }

   /* 
    * The port number can be specified in the config file (UFDBglobalPortNum is set by the parser)
    * or a command line parameter (portNumCmdLine is set in main).
    * The command line setting has preference.
    */
   if (portNumCmdLine >= 0)
      UFDBglobalPortNum = portNumCmdLine;

#if HAVE_UNIX_SOCKETS
   if (protocol == AF_UNIX)
   {
      sprintf( addr_un.sun_path, "/tmp/ufdbguardd-%05d", UFDBglobalPortNum );
#if 0
      if (unlink( addr_un.sun_path ) < 0  &&  errno != ENOENT)
         ufdbLogError( "cannot remove socket %s: %s", addr_un.sun_path, strerror(errno) );
#endif
      addr_un.sun_family = AF_UNIX;
      retval = bind( s, (struct sockaddr *) &addr_un, sizeof(addr_un) );
      /* allow anybody to connect to the socket */
      if (retval >= 0)
      {
         chmod( addr_un.sun_path, S_IRUSR|S_IWUSR| S_IRGRP|S_IWGRP| S_IROTH|S_IWOTH );
	 ufdbLogMessage( "unix socket \"%s\" successfully created", addr_un.sun_path );
      }
   }
   else
#endif
   {
      /*
       * Allow server-side addresses to be reused (don't have to wait for timeout).
       */
      sock_parm = 1;
      setsockopt( s, SOL_SOCKET, SO_REUSEADDR, (void *) &sock_parm, sizeof(sock_parm) );

      addr_in.sin_family = AF_INET;
      addr_in.sin_port = htons( UFDBglobalPortNum );
      if (strcmp( UFDBglobalInterface, "all" ) == 0)
	 addr_in.sin_addr.s_addr = htonl( INADDR_ANY );
      else
      {
	 struct in_addr iaddr;
	 if (inet_aton( UFDBglobalInterface, &iaddr ) == 0)
	 {
	    addr_in.sin_addr.s_addr = htonl( INADDR_ANY );
	    ufdbLogError( "interface parameter '%s' is invalid. I will listen on port %d on ALL interfaces.",
			  UFDBglobalInterface, UFDBglobalPortNum );
	 }
	 else
	    addr_in.sin_addr.s_addr = iaddr.s_addr;
      }
      retval = bind( s, (struct sockaddr *) &addr_in, sizeof(addr_in) );
      if (retval >= 0)
	 ufdbLogMessage( "IP socket port %d on interface \"%s\" successfully created", 
	                 UFDBglobalPortNum, UFDBglobalInterface );
   }

   if (retval < 0)
   {
      ufdbLogFatalError( "cannot bind daemon socket: %s (protocol=%s)", strerror(errno), 
                         protocol==AF_INET?"IP":"UNIX" );
      fprintf( stderr, "cannot bind daemon socket: %s (protocol=%s) *****\n", strerror(errno), 
               protocol==AF_INET?"IP":"UNIX" );
      ufdbLogMessage( "check for and kill old daemon processes" );
      fprintf( stderr, "check for and kill old daemon processes\n" );
#if HAVE_UNIX_SOCKETS
      ufdbLogMessage( "and remove UNIX socket file \"%s\"", addr_un.sun_path );
      fprintf( stderr, "and remove UNIX socket file \"%s\"\n", addr_un.sun_path );
#endif
      close( s );
      UFDBchangeStatus( UFDB_API_STATUS_FATAL_ERROR );
      pthread_exit( (void *) 0 );
   }
   
   /*
    * According to comment in the Apache httpd source code, these socket
    * options should only be set after a successful bind....
    */
   sock_parm = 1;
   setsockopt( s, SOL_SOCKET, SO_KEEPALIVE, (void *) &sock_parm, sizeof(sock_parm) );

   if (protocol == AF_INET)
   {
      sock_parm = 1;
      setsockopt( s, IPPROTO_TCP, TCP_NODELAY, (void *) &sock_parm, sizeof(sock_parm) );
   }

   if (listen( s, UFDB_MAX_WORKERS ) < 0)
   {
      ufdbLogError( "cannot listen on daemon socket: %s", strerror(errno) );
      fprintf( stderr, "cannot listen on daemon socket: %s\n", strerror(errno) );
      close( s );
      pthread_exit( (void *) 0 );
   }

   if (protocol == AF_INET)
      ufdbLogMessage( "listening on interface %s port %d", UFDBglobalInterface, UFDBglobalPortNum );
#if HAVE_UNIX_SOCKETS
   else
      ufdbLogMessage( "listening on UNIX socket \"%s\"", addr_un.sun_path );
#endif

   /* Now is the right moment to write out main PID to the pid file
    * that is used by /etc/init.d/ufdb to send signals to ufdbguardd.
    */
   writePidFile();
   atexit( removePidFile );

#ifdef _POSIX_PRIORITY_SCHEDULING
   ufdbLogMessage( "processor yielding is enabled" );
#endif
   ufdbGetSysInfo( &sysinfo );
   num_cpus = ufdbGetNumCPUs();
   ufdbLogMessage( "system: %s %s %s %s on %ld CPU%s", sysinfo.machine, sysinfo.sysname, sysinfo.release, 
                   sysinfo.nodename, num_cpus, num_cpus>1?"s":"" );
   ufdbLogMessage( "ufdbguardd " VERSION " started with %d URL verification threads and %d SSL verification threads", 
                   n_workers, UFDB_NUM_HTTPS_VERIFIERS );
   if (UFDBglobalHttpdPort > 0)
      startHttpDaemon();

   UFDBglobalReconfig = 0;

   daemon_accept_connections( s, protocol );

#if 0
   removePidFile();
#endif
   pthread_exit( NULL );

   return NULL;
}


/*
 * This thread waits and handles the following signals:
 *    HUP, CHLD, USR1, USR2, TERM, INT and ALRM.
 */
void * signal_handler_thread( void * ptr )
{
   sigset_t        sigset;
   int             sig;

   ufdbSetThreadCPUaffinity( 0 );

   /* The HUP and other signals must be blocked.
    * This is a requirement to use sigwait() in a thread.
    */
   sigemptyset( &sigset );
   sigaddset( &sigset, SIGHUP );
   sigaddset( &sigset, SIGUSR1 );
   sigaddset( &sigset, SIGUSR2 );
   sigaddset( &sigset, SIGCHLD );
   sigaddset( &sigset, SIGTERM );
   sigaddset( &sigset, SIGINT );
   sigaddset( &sigset, SIGALRM );
#if 0
   sigaddset( &sigset, SIGSEGV );	/* blocking SEGV with pthread_sigmask has undefined results */
   sigaddset( &sigset, SIGILL );
   sigaddset( &sigset, SIGBUS );
#endif
   pthread_sigmask( SIG_BLOCK, &sigset, NULL );

   while (1)
   {
      int        status;
      pid_t      pid;

      while ((pid = waitpid( -1, &status, WNOHANG )) > 0)
      {
	 if (UFDBglobalDebug > 1)
	    ufdbLogMessage( "signal handler: child process %d terminated with status %d", pid, status );
      }

      sig = 0;
      if (0 != sigwait( &sigset, &sig ))
      {
         ufdbLogError( "signal_handler_thread: sigwait() failed with %s", strerror(errno) );
	 continue;
      }

      switch (sig)
      {
      case SIGCHLD:   
         break;
      case SIGTERM:
      case SIGINT:
         TermSignalCaught( sig );
	 break;
      case SIGHUP:
	 HupSignalCaught( sig );
	 break;
      case SIGUSR1:
	 Usr1SignalCaught( sig );
         break;
      case SIGUSR2:
	 Usr2SignalCaught( sig );
         break;
      case SIGALRM:
#ifdef UFDB_HAVE_NATIVE_RWLOCK_MUTEX
	 (void) pthread_rwlock_rdlock( &TheDatabaseLock );
#endif
	 ufdbHandleAlarmForTimeEvents( UFDB_PARAM_ALARM );
#ifdef UFDB_HAVE_NATIVE_RWLOCK_MUTEX
	 (void) pthread_rwlock_unlock( &TheDatabaseLock );
#endif
         break;
      default:
         ufdbLogError( "signal handler: unexpected signal %d received", sig );
      }
   }

   return NULL;
}


/*
 * A pseudo-random sample of URLs is checked against all known categories
 * and if it appears to be uncategorised, it is buffered to be sent
 * to URLfilterDB at a later date (when SIGHUP is received).
 *
 * return 0 iff the URL is uncategorised, 1 otherwise.
 */
int ufdbVerifyURLallCats( 
   UFDBrevURL *          revURL,
   char *                URL )
{
   struct Category *     cat;
   struct UFDBmemTable * mt;

   if (UFDBglobalReconfig)
      return 1;

   if (UFDBglobalCheckedDB.table.nNextLevels > 0)
   {
      if (UFDBlookupRevUrl( &(UFDBglobalCheckedDB.table.nextLevel[0]), revURL ))
         return 1;
   }
   if (UFDBglobalCheckedExpressions != NULL)
   {
      if (ufdbRegExpMatch( UFDBglobalCheckedExpressions, URL ) == UFDB_API_MATCH)
         return 1;
   }

   for (cat = Dest;  cat != NULL;  cat = cat->next)
   {
      if (UFDBglobalReconfig)
         return 1;
      if (cat->domainlistDb != NULL)
      {
	 mt = (struct UFDBmemTable *) cat->domainlistDb->dbcp;
	 if (UFDBlookupRevUrl( &(mt->table.nextLevel[0]), revURL ))
	    return 1;
      }
      if (cat->regExp != NULL)
      {
	 if (ufdbRegExpMatch( cat->regExp, URL ) == UFDB_API_MATCH)
	    return 1;
      }
   }

   return 0;
}


static volatile int sample = 1;


static char * _blockReason(
   int reason )
{
   switch (reason)
   { 
   case UFDB_API_BLOCK_NONE:          return "PASS";
   case UFDB_API_BLOCK_ACL:           return "BLOCK";
   case UFDB_API_BLOCK_HTTPS_OPTION:  return "BLOCK";
   case UFDB_API_BLOCK_SKYPE:         return "BLOCK";
   case UFDB_API_BLOCK_SAFESARCH:     return "REDIR";
   case UFDB_API_BLOCK_LOADING_DB:    return "BLOCK-LD";
   case UFDB_API_BLOCK_FATAL_ERROR:   return "BLOCK-FATAL";
   }
   return "BLOCK";
}


/*
 * main of the worker threads.
 */
static void * worker_main( void * ptr )
{
   sigset_t                sigset;
   UFDBthreadAdmin *       admin;
   int                     tnum;
   int                     fd;
   int                     nbytes;
   unsigned long           num_reqs;
   struct Source *         src;
   struct Acl *            acl;
   char * 	 	   redirect;
   struct SquidInfo        squidInfo;
   char *                  line;
   char *                  answer;
   char *                  tmp;
   int                     reconfiguring;
#ifdef UFDB_HAVE_NATIVE_RWLOCK_MUTEX
   int                     ret;
   int                     lock_used = 0;
#endif

   tnum = (int) ((long) ptr);

   /* The HUP signal must be blocked.
    * This is a requirement to use sigwait() in a thread.
    */
   sigemptyset( &sigset );
   sigaddset( &sigset, SIGHUP );
   sigaddset( &sigset, SIGUSR1 );
   sigaddset( &sigset, SIGUSR2 );
   sigaddset( &sigset, SIGCHLD );
   sigaddset( &sigset, SIGTERM );
   sigaddset( &sigset, SIGINT );
   sigaddset( &sigset, SIGALRM );
#if 0
   sigaddset( &sigset, SIGSEGV );	/* blocking SEGV with pthread_sigmask has undefined results */
   sigaddset( &sigset, SIGILL );
   sigaddset( &sigset, SIGBUS );
#endif
   pthread_sigmask( SIG_BLOCK, &sigset, NULL );

   ufdbSetThreadCPUaffinity( tnum );

   admin = UFDBallocThreadAdmin();

   line = ufdbCalloc( 1, UFDB_MAX_URL_LENGTH );
   answer = ufdbCalloc( 1, UFDB_MAX_URL_LENGTH );
   tmp = ufdbCalloc( 1, UFDB_MAX_URL_LENGTH );

   while (1)
   {
      if (UFDBglobalTerminating)
         break;

      pthread_testcancel();
      getFdQueue( &fd );
      num_reqs = 0UL;
      if (UFDBglobalDebug)
	 ufdbLogMessage( "W%02d: open fd %2d", tnum, fd );

      while ((nbytes = read( fd, line, UFDB_MAX_URL_LENGTH-2 )) > 0)
      {
	 int access;

	 pthread_testcancel();

         line[nbytes] = '\0';
         num_reqs++;
	 UFDBlookupCounter++;	/* no mutex: faster and we do not care about a few increments less */

         if (UFDBglobalDebug > 1)
	    ufdbLogMessage( "W%02d: line=%s", tnum, line );

	 reconfiguring = UFDBglobalReconfig;
	 if (UFDBglobalTerminating)
	    break;

	 if (strncmp( line, "http://qdaemonstatus.urlfilterdb.com", 36 ) == 0)
	 {
	    int nc;
	    nc = sprintf( tmp, "http://cgibin.urlfilterdb.com/cgi-bin/URLblocked.cgi?reconfig=%d\n", reconfiguring );
	    if (write( fd, tmp, nc ) < 0)
	    {
	       ufdbLogError( "W%02d: write failed: fd=%d %s", tnum, fd, strerror(errno) );
	       goto write_error;
	    }
	    continue;
	 }

	 /* The strstr() is to prevent loops with redirect 302:... ads and other URLs
	  * that are matched by regular expressions 
	  */
	 if (strstr( line, "/URLblocked.cgi" ) != NULL  ||
	     strncmp( line, "blockedhttps.urlfilterdb.com", 28 ) == 0  ||		/* never block this URL */
	     strncmp( line, "http://cgibin.urlfilterdb.com/", 30 ) == 0)		/* never block this URL */
	 {
	    if (write( fd, "\n", 1 ) < 0)
	    {
	       ufdbLogError( "W%02d: write failed: fd=%d %s", tnum, fd, strerror(errno) );
	       goto write_error;
	    }
	    continue;
	 }

	 squidInfo.revUrl = NULL;
	 if (parseLine(admin, line, &squidInfo) != 1) 		/* break down input line into struct squidInfo */
	 {
	    line[UFDB_MAX_URL_LENGTH-2] = '\0';
#if 0
	    for (tmp = line; *tmp != '\0'; tmp++)
	       if (!isprint(*tmp))
	          *tmp = '_';
#endif
	    ufdbLogError( "W%02d: error parsing input line: %s", tnum, line );
	    if (write( fd, "\n", 1 ) < 0)
	    {
	       ufdbLogError( "W%02d: write failed: fd=%d %s", tnum, fd, strerror(errno) );
	       UFDBfreeRevURL( admin, squidInfo.revUrl );
	       goto write_error;
	    }
	    continue;
	 }

	 if (squidInfo.srcDomain[0] == '\0') 
	 {
	    squidInfo.srcDomain[0] = '-';
	    squidInfo.srcDomain[1] = '\0';
	 }

	 if (squidInfo.ident[0] == '\0') 
	 {
	    squidInfo.ident[0] = '-';
	    squidInfo.ident[1] = '\0';
	 }

	 if (strcmp( squidInfo.protocol, "https" ) == 0)
	    UFDBlookupHttpsCounter++;

	 UFDBregisterCountedIP( squidInfo.src );

	 src = NULL;

#ifdef UFDB_HAVE_NATIVE_RWLOCK_MUTEX
	 if (reconfiguring)
	 {
	    lock_used = 0;
	 }
	 else
	 {
	    lock_used = 1;
	    ret = pthread_rwlock_rdlock( &TheDatabaseLock );
	    if (ret != 0)
	       ufdbLogError( "worker_main: pthread_rwlock_rdlock failed with code %d", ret );
#ifdef UFDB_DEBUG_RWLOCK
	    else if (UFDBglobalDebug)
	       ufdbLogMessage( "worker-%02d: rdlock acquired", tnum );
#endif
	    src = UFDBgetSourceList();
	 }
#endif

	 if (strcmp( squidInfo.url, "debug.urlfilterdb.com/debug-thread-crash" ) == 0  && 
	     strcmp( squidInfo.ident, "eviltest" ) == 0)
	 {
	    struct Source * s0;
	    /* test/debug crash handler and ufdb-pstack */
	    s0 = NULL;
	    ufdbLogMessage( "worker-%02d: before forced SEGV", tnum );
	    s0->name = "crash-here-with SEGV-signal";
	    ufdbLogError( "worker-%02d: after forced SEGV", tnum );
	 }

	 /* TODO: minor optimisation: always allow HEAD */

	 for ( ; ; )
	 {
	    char * categoryIdent;
	    char * ACLident;

	    strcpy( tmp, squidInfo.src );			/* TO-DO: find out if strcpy() is necessary... */
	    src = UFDBfindSource( src, tmp, &squidInfo );
	    acl = UFDBfindACLbySource( src );

	    if (parseOnly)
	    {
	       if (write( fd, "\n", 1 ) < 0)
	       {
		  ufdbLogError( "W%02d: write failed: fd=%d %s", tnum, fd, strerror(errno) );
		  UFDBfreeRevURL( admin, squidInfo.revUrl );
		  goto write_error;
	       }
	       break;
	    }

	    redirect = tmp;
	    access = UFDBCheckACLaccess( reconfiguring, acl, &squidInfo, redirect );	/* DUNNO/ALLOW/BLOCK */

	    if (UFDBglobalDebugRedirect)
	    {
	       ufdbLogMessage( "W%02d:   REDIRECT %s %s/%s %s %s %s", tnum,
			       redirect, squidInfo.src, squidInfo.srcDomain, 
			       squidInfo.ident, "GET", squidInfo.method );
	    }

	    if (squidInfo.blockReason == UFDB_API_BLOCK_FATAL_ERROR)
	    {
	       ACLident = "config";
	       categoryIdent = "fatal-error";
	    }
	    else if (squidInfo.blockReason == UFDB_API_BLOCK_LOADING_DB)
	    {
	       ACLident = "config";
	       categoryIdent = "loading-database";
	    }
	    else 
	    {
	       if (acl == NULL)
		  ACLident = "acl-null";
	       else
		  ACLident = acl->name;

	       if (UFDBglobalFatalError)
		  categoryIdent = "fatal-error";
	       else if (reconfiguring)
		  categoryIdent = "loading-database";

	       if (squidInfo.aclpass == NULL)
	          categoryIdent = "any";
	       else if (squidInfo.aclpass->name == NULL)
		  categoryIdent = "acl-name-null";
	       else 
		  categoryIdent = squidInfo.aclpass->name;
	    }

	    if (access == UFDB_ACL_ACCESS_DUNNO  ||  access == UFDB_ACL_ACCESS_ALLOW)
	    {
	       if (src == NULL || src->cont_search == 0) 		/* sources are exhausted */
	       {
								     	/* global SafeSearch */
		  if (access == UFDB_ACL_ACCESS_DUNNO  &&
		      UFDBglobalSafeSearch  &&
		      UFDBaddSafeSearch( squidInfo.domain, squidInfo.surl, squidInfo.orig ) 
			 == UFDB_API_MODIFIED_FOR_SAFESEARCH)
		  {
		     int len;

		     UFDBsafesearchCounter++;
		     if (UFDBglobalLogAllRequests || UFDBglobalLogPass || UFDBglobalDebugRedirect)
		     {
			ufdbLogMessage( "REDIR %-11s %-15s %-10s %-9s %s %s", 
					squidInfo.ident, squidInfo.src, ACLident,
					"SAFESEARCH", squidInfo.url2display, squidInfo.method );
		     }

		     /* send a REDIRECT to Squid */
		     len = strlen( squidInfo.orig );
		     squidInfo.orig[len] = '\n';
		     len++;
		     squidInfo.orig[len] = '\0';
		     if (write( fd, squidInfo.orig, len ) < 0)
		     {
			ufdbLogError( "W%02d: write failed: fd=%d %s", tnum, fd, strerror(errno) );
			UFDBfreeRevURL( admin, squidInfo.revUrl );
			goto write_error;
		     }
		     break;
		  }

		  if (access == UFDB_ACL_ACCESS_DUNNO  &&
		      UFDBglobalYoutubeEdufilter  &&
		      UFDBaddYoutubeEdufilter( squidInfo.domain, squidInfo.surl, squidInfo.orig )
		         == UFDB_API_MODIFIED_FOR_YOUTUBE_EDUFILTER)
		  {
		     int len;

		     UFDBedufilterCounter++;
		     if (UFDBglobalLogAllRequests || UFDBglobalLogPass || UFDBglobalDebugRedirect)
		     {
			ufdbLogMessage( "REDIR %-11s %-15s %-10s %-9s %s %s",
					squidInfo.ident, squidInfo.src, ACLident,
					"EDUFILTER", squidInfo.url2display, squidInfo.method );
		     }

		     /* send a REDIRECT to Squid */
		     len = strlen( squidInfo.orig );
		     squidInfo.orig[len] = '\n';
		     len++;
		     squidInfo.orig[len] = '\0';
		     if (write( fd, squidInfo.orig, len ) < 0)
		     {
			ufdbLogError( "W%02d: write failed: fd=%d %s", tnum, fd, strerror(errno) );
			UFDBfreeRevURL( admin, squidInfo.revUrl );
			goto write_error;
		     }
		     break;
		  }

		  if (write( fd, "\n", 1 ) < 0)
		  {
		     ufdbLogError( "W%02d: write failed: fd=%d %s", tnum, fd, strerror(errno) );
		     UFDBfreeRevURL( admin, squidInfo.revUrl );
		     goto write_error;
		  }

		  if (UFDBglobalLogAllRequests ||
		      (UFDBglobalLogPass  &&  strcmp( categoryIdent, "any" ) != 0))
		  {
		     ufdbLogMessage( "PASS  %-11s %-15s %-10s %-9s %s %s",
				     squidInfo.ident, squidInfo.src, ACLident, categoryIdent, 
				     squidInfo.url2display, squidInfo.method );
		  }

		  if (!reconfiguring  &&  access == UFDB_ACL_ACCESS_DUNNO)
		  {
		     if ((squidInfo.domain[0] <= '9'  &&  squidInfo.domain[0] >= '0')  &&
			 ( (strncmp( squidInfo.domain, "10.", 3 )==0) ||
			   (strncmp( squidInfo.domain, "127.", 4 )==0) ||
			   (strncmp( squidInfo.domain, "169.254.", 8 )==0) ||
			   (strncmp( squidInfo.domain, "192.168.", 8 )==0) ))
		     {
			/* do NOT check 10.*, 127.* and 192.168.*  with uncategorised domains */
		     }
		     else
		     {
			if (!ufdbVerifyURLallCats( squidInfo.revUrl, squidInfo.url2display ))
			{
			   ufdbRegisterUnknownURL( squidInfo.domain, squidInfo.port );
			   UFDBuncategorisedCounter++;
			}
		     }
		  }

		  break;
	       }

	       /* source != NULL */

	       if (access == UFDB_ACL_ACCESS_DUNNO  &&  src->next != NULL) 
	       {
		  src = src->next;
		  continue;
	       }
	       else
	       {
		  if (write( fd, "\n", 1 ) < 0)
		  {
		     ufdbLogError( "W%02d: write failed: fd=%d %s", tnum, fd, strerror(errno) );
		     UFDBfreeRevURL( admin, squidInfo.revUrl );
		     goto write_error;
		  }
		  break;
	       }
	    }
	    else			/* access == UFDB_ACL_ACCESS_BLOCK */
	    {
	       if (testMode)
	       {
		  ufdbLogMessage( "TEST-%-5s %s %s %s %s %s %s",
				  _blockReason(squidInfo.blockReason),
				  squidInfo.ident, squidInfo.src, ACLident,
				  categoryIdent, squidInfo.url2display, squidInfo.method );
		  if (write( fd, "\n", 1 ) < 0)	/* send back an "OK" message */
		  {
		     ufdbLogError( "W%02d: write failed: fd=%d %s", tnum, fd, strerror(errno) );
		     UFDBfreeRevURL( admin, squidInfo.revUrl );
		     goto write_error;
		  }
		  UFDBtestBlockCounter++;
	       }
	       else
	       {
		  char urlgrp[64];

		  if (squidInfo.urlgroup[0] == '#')
		     urlgrp[0] = '\0';
		  else
		  {
		     urlgrp[0] = ' ';
		     strncpy( &urlgrp[1], squidInfo.urlgroup, 62 );
		     urlgrp[63] = '\0';
		  }

		  if (UFDBglobalLogBlock || UFDBglobalLogAllRequests)
		     ufdbLogMessage( "%-5s %-11s %-15s %-10s %-9s %s %s",
				     _blockReason(squidInfo.blockReason),
				     squidInfo.ident, squidInfo.src, ACLident,
				     categoryIdent, squidInfo.url2display, squidInfo.method );
		  sprintf( answer, "%s %s/%s %s %s%s\n", redirect, squidInfo.src,
			   squidInfo.srcDomain, squidInfo.ident, "GET", urlgrp );
		  if (write( fd, answer, strlen(answer) ) < 0)
		  {
		     ufdbLogError( "W%02d: write failed: fd=%d %s", tnum, fd, strerror(errno) );
		     UFDBfreeRevURL( admin, squidInfo.revUrl );
		     goto write_error;
		  }
		  if (UFDBglobalDebug > 1)
		     ufdbLogMessage( "   redirected to %s", answer );
		  UFDBblockCounter++;
	       }
	       break;
	    }
	 }

#ifdef UFDB_HAVE_NATIVE_RWLOCK_MUTEX
	 if (lock_used)
	 {
	    ret = pthread_rwlock_unlock( &TheDatabaseLock );
	    if (ret != 0)
	       ufdbLogError( "worker-%2d: error releasing lock: %d", tnum, ret );
#ifdef UFDB_DEBUG_RWLOCK
	    else if (UFDBglobalDebug)
	       ufdbLogMessage( "worker-%02d: rdlock release", tnum );
#endif
	    lock_used = 0;
	 }
#endif

	 UFDBfreeRevURL( admin, squidInfo.revUrl );

	 /* We are slowing down all traffic by introducing an artificial 0.1 second delay.
	  * The effect of this delay is that
	  * 1) the database reload thread gets the CPU more often.
	  * 2) during the database reload only a few URLs are verified (slowly) which means
	  *    that the user experiences a sluggish internet connection and much less URLs
	  *    are passed through unfiltered.
	  */
         if (reconfiguring  &&  UFDBglobalURLlookupDelayDBreload)
	    usleep( 100000 );

#ifdef _POSIX_PRIORITY_SCHEDULING
	 if (UFDBhttpsVerificationQueued() > 3)
	    sched_yield();	
#endif
	 if (sig_other)
	    break;
      }				/* end of while-read loop */

      if (nbytes < 0)
      {
         ufdbLogError( "W%02d: read failed: fd=%d %s", tnum, fd, strerror(errno) );
      }
write_error:

#ifdef UFDB_HAVE_NATIVE_RWLOCK_MUTEX
      if (lock_used)
      {
	 ret = pthread_rwlock_unlock( &TheDatabaseLock );
	 if (ret != 0)
	    ufdbLogError( "worker-%2d: error releasing lock: %d", tnum, ret );
#ifdef UFDB_DEBUG_RWLOCK
	 else if (UFDBglobalDebug)
	    ufdbLogMessage( "worker-%02d: rdlock release", tnum );
#endif
	 lock_used = 0;
      }
#endif

      if (num_reqs > 1UL)
	 ufdbLogMessage( "W%02d: %lu URL verifications", tnum, num_reqs );

      if (UFDBglobalDebug)
	 ufdbLogMessage( "W%02d: close fd %2d", tnum, fd );
      close( fd );
      fd = -1;
   }

   /*NOTREACHED*/
   pthread_exit( NULL );
}


/*
 *  main of thread that deauls with source that use execuserlist.
 */
static void * dynamic_userlist_updater_main( void * ptr )
{
   sigset_t                sigset;
   struct Source *         src;
#ifdef UFDB_HAVE_NATIVE_RWLOCK_MUTEX
   int                     ret;
#endif

   sigemptyset( &sigset );
   sigaddset( &sigset, SIGHUP );
   sigaddset( &sigset, SIGUSR1 );
   sigaddset( &sigset, SIGUSR2 );
   sigaddset( &sigset, SIGTERM );
   sigaddset( &sigset, SIGINT );
   sigaddset( &sigset, SIGALRM );
#if 0
   sigaddset( &sigset, SIGCHLD );	/* SIGCHLD is used to interrupt nanosleep() */
#endif
#if 0
   sigaddset( &sigset, SIGSEGV );	/* blocking SEGV with pthread_sigmask has undefined results */
   sigaddset( &sigset, SIGILL );
   sigaddset( &sigset, SIGBUS );
#endif
   pthread_sigmask( SIG_BLOCK, &sigset, NULL );

   ufdbSetThreadCPUaffinity( 0 );

   while (1)
   {
      struct timespec  tv;

      pthread_testcancel();

      if (UFDBglobalDebug)
	 ufdbLogMessage( "dynamic_userlist_updater_main: sleeping for %d minutes", UFDBglobalRefreshUserlistInterval );

      tv.tv_sec = UFDBglobalRefreshUserlistInterval * 60;
      tv.tv_nsec = 0;
      (void) nanosleep( &tv, NULL );

      if (UFDBglobalTerminating)
         break;
      if (UFDBglobalReconfig)
         continue;
      
      if (UFDBglobalDebug || UFDBglobalDebugExternalScripts)
	 ufdbLogMessage( "dynamic_userlist_updater_main: going to refresh execuserlists" );

#ifdef UFDB_HAVE_NATIVE_RWLOCK_MUTEX
      ret = pthread_rwlock_rdlock( &TheDatabaseLock );
      if (ret != 0)
	 ufdbLogError( "dynamic_userlist_updater_main: pthread_rwlock_rdlock TheDatabaseLock failed with code %d", ret );
#ifdef UFDB_DEBUG_RWLOCK
      else if (UFDBglobalDebug)
	 ufdbLogMessage( "dynamic_userlist_updater_main: rdlock TheDatabaseLock acquired" );
#endif
#endif

      src = UFDBgetSourceList();
      while (src != NULL)
      {
         if (src->active  &&  src->userDb != NULL  &&  src->userDb->type == SGDBTYPE_EXECUSERLIST)
	 {
	    struct sgDb * oldudb;
	    struct sgDb * newudb;

	    if (UFDBglobalDebug || UFDBglobalDebugExternalScripts)
	       ufdbLogMessage( "dynamic_userlist_updater_main: refreshing execuserlist for %s (%s)", src->name, src->sarg0 );

	    oldudb = src->userDb;
	    newudb = UFDBretrieveExecUserlist( src->sarg0 );
	    
#ifdef UFDB_HAVE_NATIVE_RWLOCK_MUTEX
	    ret = pthread_rwlock_wrlock( &TheDynamicSourcesLock );
	    if (ret != 0)
	       ufdbLogError( "dynamic_userlist_updater_main: pthread_rwlock_wrlock TheDynamicSourcesLock failed with code %d", ret );
#else
	    ret = pthread_mutex_lock( &TheDynamicSourcesLock );
	    if (ret != 0)
	       ufdbLogError( "dynamic_userlist_updater_main: pthread_mutex_lock TheDynamicSourcesLock failed with code %d", ret );
#endif
            src->userDb = newudb;

#ifdef UFDB_HAVE_NATIVE_RWLOCK_MUTEX
	    ret = pthread_rwlock_unlock( &TheDynamicSourcesLock );
	    if (ret != 0)
	       ufdbLogError( "dynamic_userlist_updater_main: pthread_rwlock_unlock TheDynamicSourcesLock failed with code %d", ret );
#else
	    ret = pthread_mutex_unlock( &TheDynamicSourcesLock );
	    if (ret != 0)
	       ufdbLogError( "dynamic_userlist_updater_main: pthread_mutex_unlock TheDynamicSourcesLock failed with code %d", ret );
#endif

            UFDBmemDBdeleteDB( (struct UFDBmemDB *) oldudb->dbcp );
	    ufdbFree( oldudb );
	 }
         src = src->next;
      }

#ifdef UFDB_HAVE_NATIVE_RWLOCK_MUTEX
      ret = pthread_rwlock_unlock( &TheDatabaseLock );
      if (ret != 0)
	 ufdbLogError( "dynamic_userlist_updater_main: error releasing lock: %d", ret );
#ifdef UFDB_DEBUG_RWLOCK
      else if (UFDBglobalDebug)
	 ufdbLogMessage( "dynamic_userlist_updater_main: rdlock release" );
#endif
#endif

      if (UFDBglobalDebug || UFDBglobalDebugExternalScripts)
	 ufdbLogMessage( "dynamic_userlist_updater_main: refresh completed" );
   }

   /*NOTREACHED*/
   pthread_exit( NULL );
}


void doCommand( char * command )
{
   /* There is a serious problem sending signals to ufdbguardd and ufdbhttpd.
    * 1) ufdbguardd cannot send SIGTERM, but can send SIGKILL to ufdbhttpd.
    * 2) ufdbUpdate cannot send SIGHUP to ufdbguardd.
    */

   if (strcmp( command, "verify" ) == 0)
   {
      testConfig = 1;
      runAsDaemon = 0;
      UFDBglobalUploadStats = 0;
      globalErrorLog = stderr;
   }
}


extern char ** environ;

int main( 
   int            argc,
   char **        argv )
{
   int            i;
   pid_t          pid;
   sigset_t       sigset;
   pthread_attr_t attr;
   time_t 	  t;
   void *         dummy_ptr;
   char           niso_str[22];

   char ** envp = environ;

#ifdef UFDB_HAVE_NATIVE_RWLOCK_MUTEX
   pthread_rwlockattr_t mylock_attr;
#endif

   strcpy( progname, "ufdbguardd" );

   /* TODO: unset LC variables; regex cannot deal with multibyte characters */

   while ((i = getopt(argc, argv, "AC:TNdDRPSh?rt:c:L:p:U:vw:")) != EOF)
   {
      switch (i) {
      case 'A':
	 fprintf( stderr, "-A option is obsolete\n");
	 break;
      case 'C':
	 doCommand( optarg );
         break;
      case 'N':
         UFDBglobalAnalyseUncategorisedURLs = -1;
	 break;
      case 'D':
	 runAsDaemon = 0;	/* undocumented -D option to prevent running as daemon */
	 break;
      case 'd':
	 UFDBglobalDebug++;
#ifdef UFDB_DEBUG
	 yydebug = 1;
#endif
	 break;
      case 'c':
	 configFile = optarg;
	 break;
      case 'p':
      	 portNumCmdLine = atoi( optarg );
	 if (portNumCmdLine <= 0) {
	    fprintf( stderr, "port number must be > 0\n" );
	    exit( 1 );
	 }
	 break;
      case 'P':			/* undocumented -P option for development purposes */
	 parseOnly = 1;
	 break;
      case 'R':
         UFDBglobalDebugRegexp = 1;
	 break;
      case 'S':
         UFDBglobalUploadStats = 0;
	 break;
      case 'r':
	 UFDBglobalDebugRedirect = 1;
	 break;
      case 't':			/* undocumented -t option for time-related testing */
	 if ((t = iso2sec(optarg)) == -1) {
	    fprintf( stderr, "-t dateformat error, should be yyyy-mm-ddTHH:MM:SS\n" );
	    exit( 1 );
	 }
	 if (t < 0) {
	    fprintf( stderr, "-t date have to after 1970-01-01T01:00:00\n" );
	    exit( 1 );
	 }
	 niso( t, niso_str );
	 ufdbLogMessage( "ufdbguardd emulating date %s", niso_str );
	 globalDebugTimeDelta = t - start_time.tv_sec;
	 start_time.tv_sec = start_time.tv_sec + globalDebugTimeDelta;
	 break;
      case 'T':
	 testMode = 1;		
	 break;
      case 'U':
	 if (strlen(optarg) <= 31)
	    strcpy( UFDBglobalUserName, optarg );
	 else
	    ufdbLogFatalError( "username supplied with -U option is too long" );
         break;
      case 'L':				/* undocumented -L option for alternate PID file */
	 globalPidFile = optarg;
         break;
      case 'v':
	 fprintf( stderr, "ufdbguardd: %s\n", VERSION );
	 fprintf( stderr, "The ufdbGuard software suite is free and Open Source Software.\n" );
	 fprintf( stderr, "Copyright (C) 2005-2012 by URLfilterDB and others.\n" );
	 exit( 0 );
	 break;
      case 'w':
         n_workers = atoi( optarg );
	 if (n_workers < 32)
	    n_workers = 32;
	 else if (n_workers > UFDB_MAX_WORKERS)
	 {
	    fprintf( stderr, "-w option exceeds the maximum.  The #threads is set to %d\n", UFDB_MAX_WORKERS );
	    n_workers = UFDB_MAX_WORKERS;
	 }
	 break;
      case '?':
      case 'h':
         usage( '\0' );
	 break;
      default:
         usage( i );
      }
   }

   if (runAsDaemon)
   {
      if ((pid = fork()) != 0)
	 exit( 0 );        
#ifndef UFDB_DEBUG
      close( 0 );
      close( 1 );
#endif
      setsid();
   }
   globalPid = getpid();

   UFDBdropPrivileges( UFDBglobalUserName );

   UFDBinitializeIPcounters();

   /*
    * Initialise signal handlers.
    * The HUP signal is dealt with on a per thread basis.
    */
   ufdbSetSignalHandler( SIGILL,  ufdbCatchBadSignal );
   ufdbSetSignalHandler( SIGBUS,  ufdbCatchBadSignal );
   ufdbSetSignalHandler( SIGSEGV, ufdbCatchBadSignal );

   ufdbSetSignalHandler( SIGPIPE, SIG_IGN );

   ufdbSetSignalHandler( SIGABRT, catchAbortSignal );

   ufdbSetSignalHandler( SIGTERM, catchTermSignal );
   ufdbSetSignalHandler( SIGINT,  catchTermSignal );

   ufdbSetSignalHandler( SIGHUP,  catchHUPSignal );
   ufdbSetSignalHandler( SIGUSR1, catchUSR1Signal );
   ufdbSetSignalHandler( SIGUSR2, catchUSR2Signal );

   ufdbSetSignalHandler( SIGCHLD, catchChildSignal );

   /* All signals must be blocked.
    * This is a requirement to use sigwait() in a thread.
    */
   sigemptyset( &sigset );
   sigaddset( &sigset, SIGHUP );
   sigaddset( &sigset, SIGUSR1 );
   sigaddset( &sigset, SIGUSR2 );
   sigaddset( &sigset, SIGCHLD );
   sigaddset( &sigset, SIGTERM );
   sigaddset( &sigset, SIGINT );
   sigaddset( &sigset, SIGALRM );
#if 0
   sigaddset( &sigset, SIGILL );
   sigaddset( &sigset, SIGBUS );
   sigaddset( &sigset, SIGSEGV );	/* blocking SEGV with pthread_sigmask has undefined results */
#endif
   pthread_sigmask( SIG_BLOCK, &sigset, NULL );

   my_argc = argc;
   my_argv = argv;
   my_envp = envp;

   if (testConfig)
   {
      UFDBglobalFatalError = 0;
      UFDBglobalHttpdPort = 0;
      sgReadConfig( configFile );
      if (UFDBglobalFatalError)
      {
	 ufdbLogMessage( "A FATAL ERROR OCCURRED: (see previous lines for more details)  *****" );
      }
      exit( 0 );
   }

#ifdef UFDB_HAVE_NATIVE_RWLOCK_MUTEX
   pthread_rwlockattr_init( &mylock_attr );
#if HAVE_PTHREAD_RWLOCK_PREFER_WRITER_NP
   pthread_rwlockattr_setkind_np( &mylock_attr, PTHREAD_RWLOCK_PREFER_WRITER_NP );
#endif
   pthread_rwlock_init( &TheDatabaseLock, &mylock_attr );
#endif

   /*
    * create the threads.
    */
   pthread_attr_init( &attr );
   pthread_attr_setscope( &attr, PTHREAD_SCOPE_SYSTEM );
   pthread_attr_setstacksize( &attr, 256 * 1024 );

   pthread_create( &threadedSignalHandler, &attr, signal_handler_thread, (void *) 0 );
#ifdef _POSIX_PRIORITY_SCHEDULING
   sched_yield();
#else
   usleep( 100000 );
#endif

   pthread_create( &sockethandler, &attr, socket_handler_thread, (void *) 0 );
   usleep( 200000 );

   pthread_create( &dyn_userlist_handler, &attr, dynamic_userlist_updater_main, (void *) 0 );
   usleep( 200000 );

   for (i = 0; i < n_workers; i++)
   {
      pthread_create( &workers[i], &attr, worker_main, (void *) ((long) i) );
#ifdef _POSIX_PRIORITY_SCHEDULING
      sched_yield();
#else
      usleep( 15000 );
#endif
   }
   if (UFDBglobalDebug)
      ufdbLogMessage( "using %d worker threads", n_workers );

   /*
    * exit when the sockethandler thread terminates.
    */
   pthread_join( sockethandler, &dummy_ptr );

   if (UFDBglobalDebug)
      ufdbLogMessage( "Socket handler stopped. ufdbguardd is going to stop." );

   /*
    *  Note: sockethandler terminated because of a signal.
    */
   UFDBglobalReconfig = 1;
   UFDBglobalTerminating = 1;
   alarm( 0 );

   if (badSignal)
      BadSignalCaught( badSignal );

   /* cancel (soft termination) and kill (hard and quick termination) threads */
   if (UFDBglobalDebug)
      ufdbLogMessage( "main: going to cancel threads" );
   pthread_cancel( dyn_userlist_handler );
   for (i = 0; i < UFDB_NUM_HTTPS_VERIFIERS; i++)
      pthread_cancel( httpsthr[i] );
   for (i = 0; i < n_workers; i++)
      pthread_cancel( workers[i] );
   usleep( 500000 );
   pthread_cancel( threadedSignalHandler );

   removePidFile();
   UFDBchangeStatus( UFDB_API_STATUS_TERMINATED );
   ufdbLogMessage( "ufdbGuard daemon stopped" );

#if defined(UFDB_FREE_MEMORY)
   ufdbLogMessage( "freeing all memory ..." );
   ufdbFreeAllMemory();
   ufdbLogMessage( "done freeing memory." );
#endif

   _exit( 0 );

   return 0;
}


static void uploadStatistics( 
   char * reason )
{
   char * URLs;
   char * message;
   int    length;
   int    s;
   int    nbytes;
   int    written;
   int    dummy;
   int    status;
   long   num_cpus;
   unsigned long  num_clients;
   struct utsname sysinfo;

   num_clients = UFDBgetNumberOfRegisteredIPs();
   if (UFDBglobalCheckedDB.mem == NULL)		/* not a client of URLfilterDB */
   {
      if (UFDBlookupCounter < 200)
	 return;
      if (num_clients < 2)
         return;
   }

   if (UFDBglobalDebug)
      ufdbLogMessage( "uploadStatistics: upload uncategorised URLs: %s,  upload statistics: %s", 
                      UFDBglobalAnalyseUncategorisedURLs>0 ? "yes" : "no",
		      UFDBglobalUploadStats ? "yes" : "no" );

   if (UFDBglobalAnalyseUncategorisedURLs > 0)
   {
      URLs = ufdbGetUnknownURLs();
      if (URLs == NULL)
	 URLs = "example.com|";
   }
   else
      URLs = "example.com|";

   if (UFDBglobalLogUncategorisedURLs)
   {
      if (!UFDBglobalAnalyseUncategorisedURLs)
	 ufdbLogError( "log-uncategorised-urls is ON but analyse-uncategorised-urls is OFF and therefore there are no uncategorised URLs to be logged" );
      else if (UFDBuncategorisedCounter == 0)
      {
         ufdbLogMessage( "there are no uncategorised URLs" );
      }
      else
      {
	 char * u;
         for (u = URLs; *u != '\0'; )
	 {
	    char * sep;
	    int    len;
	    sep = strchr( u, '|' );
	    if (sep == NULL)
	       break;
	    len = (int) (sep - u);
	    ufdbLogMessage( "uncategorised URL: %*.*s", len, len, u );
	    u = sep + 1;
	 }
      }
   }

   length = strlen( URLs );
   if (length <= 1)
   {
      URLs = "example.com|";
      length = strlen( URLs );
   }

   message = ufdbMalloc( 1500 + length );
   if (message == NULL)
   {
      ufdbLogError( "could not malloc %d bytes for upload message. Aborting upload of uncategorised URLs", 1500 + length );
      ufdbResetUnknownURLs();
      return;
   }

   ufdbGetSysInfo( &sysinfo );
   num_cpus = ufdbGetNumCPUs();

   if (UFDBglobalUploadStats)
   {
      sprintf( message, 
	       "POST /cgi-bin/uncat.pl HTTP/1.1\r\n"
	       "Host: " UFDB_UPLOAD_UNCATEGORISED_URLS_WEBSITE "\r\n"
	       "User-Agent: ufdbGuardd-" VERSION "\r\n"
	       "Content-Type: text/plain\r\n"
	       "Content-Length: %d\r\n"
	       "Connection: close\r\n"
	       "X-HasURLfilterDB: %s\r\n"
	       "X-SiteInfo: %s %ld %s %s\r\n"
	       "X-NodeName: %s\r\n"
	       "X-NumClients: %lu\r\n"
	       "X-NumLookups: %lu\r\n"
	       "X-NumHttpsLookups: %lu\r\n"
	       "X-NumTunnelsDetected: %lu\r\n"
	       "X-NumTestBlock: %lu\r\n"
	       "X-NumBlock: %lu\r\n"
	       "X-NumSafeSearch: %lu\r\n"
	       "X-Uncategorised: %lu\r\n"
	       "X-UploadSeqNo: %lu\r\n"
	       "X-UploadReason: %s\r\n"
	       "\r\n"
	       "%s\r\n",
	       length, 
	       (UFDBglobalCheckedDB.mem == NULL ? "no" : "yes"),
	       sysinfo.machine, num_cpus, sysinfo.sysname, sysinfo.release, 
	       sysinfo.nodename,
	       num_clients,
	       UFDBlookupCounter,
	       UFDBlookupHttpsCounter,
	       UFDBglobalTunnelCounter,
	       UFDBtestBlockCounter,
	       UFDBblockCounter,
	       UFDBsafesearchCounter,
	       UFDBuncategorisedCounter,
	       UFDBuploadSeqNo,
	       reason,
	       URLs );
   }
   else
   {
      sprintf( message, 
	       "POST /cgi-bin/uncat.pl HTTP/1.1\r\n"
	       "Host: " UFDB_UPLOAD_UNCATEGORISED_URLS_WEBSITE "\r\n"
	       "User-Agent: ufdbGuardd-" VERSION "\r\n"
	       "Content-Type: text/plain\r\n"
	       "Content-Length: %d\r\n"
	       "X-HasURLfilterDB: %s\r\n"
	       "X-NodeName: %s\r\n"
	       "X-NumClients: %lu\r\n"
	       "\r\n",
	       length,
	       (UFDBglobalCheckedDB.mem == NULL ? "no" : "yes"),
	       sysinfo.nodename,
	       num_clients  );
   }
   length = strlen( message );

   /* This function migth be called N times when a bad signal is received.
    * Let's try to execute the upload only once by resetting the unknown URLs quickly.
    */
   ufdbResetUnknownURLs();

   s = UFDBopenSocket( UFDB_UPLOAD_UNCATEGORISED_URLS_WEBSITE, 80 );
   if (s < 0)
   {
      ufdbLogError( "cannot upload statistics/uncategorised URLs  *****\n"
                    "cannot open a communication socket to %s:%d (%s)",
                    UFDB_UPLOAD_UNCATEGORISED_URLS_WEBSITE, 80, strerror(errno) );
      return;
   }

   written = 0;
   while (length > 0)
   {
      nbytes = write( s, message+written, length );
      if (nbytes == 0)
         break;
      if (nbytes < 0)
      {
         if (errno != EINTR)
	    break;
      }
      else
      {
         written += nbytes;
	 length -= nbytes;
      }
   }

   errno = 0;
   status = 0;
   nbytes = read( s, message, 700 );
   if (nbytes < 0 || nbytes < 9)
   {
      ufdbLogError( "did not receive a HTTP response header. URLs are not uploaded: %s *****", strerror(errno) );
      written = 0;
   }
   else if (sscanf( message, "HTTP/1.%d %d ", &dummy, &status ) != 2)
   {
      ufdbLogError( "received a malformed HTTP response header: URLs are not uploaded *****" );
      written = 0;
   }
   else if (status != 200)
   {
      ufdbLogError( "received an unexpected HTTP response status: code %d. URLs are not uploaded  *****", status );
      written = 0;
   }

   if (UFDBglobalDebug)
   {
      if (nbytes > 0)
      {
         message[nbytes] = '\0';
	 ufdbLogMessage( "URL upload received this HTTP reply:\n%s", message );
	 ufdbLogMessage( "HTTP reply status is %d  (status message has %d bytes)", status, nbytes );
      }
   }

   close( s );
   ufdbFree( message );

   if (written>0  &&  (UFDBglobalDebug  ||  UFDBglobalAnalyseUncategorisedURLs > 0))
      ufdbLogMessage( "uncategorised URLs have been uploaded to URLfilterDB" );
}

