/*
 * ufdblib.h - URLfilterDB
 *
 * ufdbGuard is copyrighted (C) 2005-2013 by URLfilterDB with all rights reserved.
 *
 * RCS $Id: ufdblib.h,v 1.71 2013/01/14 02:32:08 root Exp root $
 */

#ifndef UFDB_UFDBLIB_H_INCLUDED
#define UFDB_UFDBLIB_H_INCLUDED

#include "ufdb.h"

#include <time.h>
#include <stdio.h>
#include <sys/times.h>
#include <sys/utsname.h>


extern volatile int    UFDBglobalStatus;
extern volatile int    UFDBglobalDatabaseStatus;
extern volatile int    UFDBglobalTerminating;
extern volatile int    UFDBglobalReconfig;
extern volatile int    UFDBglobalDebug;
extern volatile int    UFDBglobalLogAllRequests;
extern volatile int    UFDBglobalLogPass;
extern volatile int    UFDBglobalLogBlock;
extern volatile int    UFDBglobalSafeSearch;
extern volatile int    UFDBglobalStripDomainFromUsername;
extern volatile int    UFDBglobalDebugRedirect;
extern volatile int    UFDBglobalUploadStats;
extern volatile int    UFDBglobalYoutubeEdufilter;

extern time_t          UFDBglobalDatabaseLoadTime;
extern int    	       UFDBglobalFatalError;
extern char            UFDBglobalLicenseStatus[512];
extern char            UFDBglobalRedirectHttps[1024];
extern char * UFDBglobalConfigFile;
extern int    UFDBglobalDebugSkype;
extern int    UFDBglobalDebugGtalk;
extern int    UFDBglobalDebugYahooMsg;
extern int    UFDBglobalDebugAim;
extern int    UFDBglobalDebugFBchat;
extern int    UFDBglobalDebugCitrixOnline;
extern int    UFDBglobalDebugHttpd;
extern int    UFDBglobalDebugRegexp;
extern int    UFDBglobalExpressionOptimisation;
extern int    UFDBglobalRefreshUserlistInterval;
extern int    UFDBglobalDebugExternalScripts;
extern char   UFDBglobalUserName[31+1];
extern char * UFDBglobalLogDir;
extern char * UFDBglobalEmailServer;
extern char * UFDBglobalAdminEmail;
extern char * UFDBglobalExternalStatusCommand;
extern char * UFDBglobalYoutubeEdufilterID;
extern int    UFDBglobalURLlookupDelayDBreload;
extern int    UFDBglobalURLlookupResultDBreload;
extern int    UFDBglobalURLlookupResultFatalError;
extern int    UFDBglobalAnalyseUncategorisedURLs;
extern int    UFDBglobalLogUncategorisedURLs;
extern int    UFDBglobalPortNum;
extern int    UFDBglobalShowURLdetails;
extern int    UFDBglobalTunnelCheckMethod;
extern int    UFDBglobalSquidUsesSSLbump;
extern int    UFDBglobalHttpsWithHostname;
extern int    UFDBglobalHttpsOfficialCertificate;
extern int    UFDBglobalHttpsEnforceContentPeek;
extern int    UFDBglobalSkypeOverHttps;
extern int    UFDBglobalGtalkOverHttps;
extern int    UFDBglobalYahooMsgOverHttps;
extern int    UFDBglobalAimOverHttps;
extern int    UFDBglobalFBchatOverHttps;
extern int    UFDBglobalCitrixOnlineOverHttps;
extern int    UFDBglobalUnknownProtocolOverHttps;
extern int    UFDBglobalHttpsNoSSLv2;
extern int    UFDBglobalHttpdPort;
extern char   UFDBglobalHttpdInterface[256];
extern char   UFDBglobalHttpdImagesDirectory[256];
extern char   UFDBglobalInterface[256];
extern char   UFDBglobalCAcertsFile[1024];
extern char   UFDBglobalCAcertsDir[1024];
extern char   UFDBglobalFatalErrorRedirect[1024];
extern char   UFDBglobalLoadingDatabaseRedirect[1024];
extern struct UFDBmemTable UFDBglobalCheckedDB;
extern struct UFDBmemTable UFDBglobalSafeDB;
extern struct ufdbRegExp * UFDBglobalCheckedExpressions;
extern unsigned long UFDBglobalMaxLogfileSize;
volatile extern unsigned long UFDBglobalTunnelCounter;

extern struct ufdbSetting * lastSetting;
extern struct ufdbSetting * Setting;

extern struct Source * lastSource;
extern struct Source * saveSource;
extern volatile struct Source * SourceList;

extern struct Category * lastDest;
extern struct Category * Dest;

extern struct sgRewrite * lastRewrite;
extern struct sgRewrite * Rewrite;
extern struct ufdbRegExp *  lastRewriteRegExec;

extern struct ufdbTime * lastTime;
extern struct ufdbTime * TimeList;

extern FILE * globalErrorLog;
extern struct LogFile * globalLogFile;

extern struct LogFileStat * lastLogFileStat;
extern struct LogFileStat * LogFileStat;

extern struct TimeElement *lastTimeElement;
extern struct TimeElement *TimeElement;

extern struct Acl * lastAcl;
extern struct Acl * defaultAcl;
extern struct AclDest * lastAclDest;
extern struct Acl * AclList;

extern struct ufdbRegExp * lastRegExpDest;

extern char ** globalArgv;
extern char ** globalEnvp;
extern int globalDebugTimeDelta;
extern int globalPid;
extern int globalUpdate;
extern char * globalCreateDb;
extern int failsafe_mode;
extern int sig_hup;
extern int sig_other;
extern int httpsConnectionCacheSize;
extern char progname[80];

void UFDBdropPrivileges( const char * username );
void UFDBraisePrivileges( const char * username, const char * why );

struct UFDBmemDB;	/* forward declaration */

struct UFDBmemDB
{
   char * key;
   char * value;
   int    length;
   struct UFDBmemDB * next;
};

struct UFDBmemDB * UFDBmemDBinit( void );
void UFDBmemDBinsert( struct UFDBmemDB * db, char * key, char * value, int length );
int UFDBmemDBfind( struct UFDBmemDB * db, char *  key, char ** value );
void UFDBmemDBdeleteDB( struct UFDBmemDB * db );
void UFDBmemDBprintUserDB( struct UFDBmemDB * db );

void UFDBappInit( void );
unsigned long UFDBappMemUsage( void );

void UFDBtimerInit( struct tms * t );
void UFDBtimerStop( struct tms * t );
void UFDBtimerPrint( struct tms * t, char * tag );
void UFDBtimerPrintString( char * line, struct tms * t, char * tag );
void UFDBlogTimer( struct tms * t, char * tag );

int UFDBloadDatabase( 
   struct UFDBmemTable * mtable, 
   char *                file    );

int UFDBloadExpressions( 
   struct ufdbRegExp ** exprList,
   char *               file     );

struct ufdbRegExp * UFDBoptimizeExprList( 
   char *               reSource,
   struct ufdbRegExp *  reList );

void ufdbFreeRegExprList( struct ufdbRegExp * re );

int  UFDBregcomp( regex_t * preg, const char * regex, int cflags );
void UFDBregfree( regex_t * preg );
int  UFDBregexec( const regex_t * preg, const char * string, size_t nmatch, regmatch_t pmatch[], int eflags );

struct Source * UFDBgetSourceList( void );
void UFDBsetSourceList( struct Source * sl );

int  ufdbSetCPU( char * CPUspec );
int  ufdbSetThreadCPUaffinity( int thread_num );
void ufdbResetCPUs( void );

char * ufdbAPIstatusString( int api_code );
char * ufdbDBstat2string( int status );
char * ufdbStatus2string( int status );

#define URL_HIST_SIZE 120000

int    ufdbVerifyURLallCats( UFDBrevURL * revURL, char * URL );
int    ufdbRegisterUnknownURL( char * webserver, int portnumber );
char * ufdbGetUnknownURLs( void );
void   ufdbResetUnknownURLs( void );

int UFDBopenSocket( char * serverName, int port );

void UFDBregisterCountedIP( const char * address );
void UFDBinitializeIPcounters( void );
unsigned long UFDBgetNumberOfRegisteredIPs( void );

struct ufdbSetting {
  char *           name;
  char *           value;
  struct ufdbSetting * next;
};

void   ufdbSetting( char *, char * );
struct ufdbSetting * ufdbSettingFindName( char * );
char * ufdbSettingGetValue( char * );

void   ufdbLogFile( int, int, char * );
void   ufdbGlobalSetLogging( int logging );
void   ufdbRotateLogfile( char * );
void   UFDBrotateLogfile( void );
void   ufdbSetGlobalErrorLogFile( void );
void   UFDBcloseFilesNoLog( void );

void   ufdbLogMessage( char * format, ... )
#ifdef __GNUC__
__attribute__ ((format(printf,1,2)))
#endif
;
void   ufdbLogError( char *, ... )
#ifdef __GNUC__
__attribute__ ((format(printf,1,2)))
#endif
;
void   ufdbLogFatalError( char *, ... )
#ifdef __GNUC__
__attribute__ ((format(printf,1,2)))
#endif
;

void ufdbGetSysInfo( struct utsname * si );
long ufdbGetNumCPUs( void );

void UFDBlogConfig( void );
int  UFDBchangeStatus( int status );

char * UFDBparseIPv6address( char * url, char * domain );
void   UFDBupdateURLwithNormalisedDomain( char * url, char * newDomain );

/*
 * strip a URL;
 * remove http:// prefix, remove www[0-9*]. prefix,
 * remove port number, remove username and password
 */
void UFDBstripURL(
   char * URL,                  /* input URL string */
   char * strippedUrl,          /* output char array (must be 1024 bytes) */
   char * domain,               /* output char array (must be 1024 bytes) */
   char * protocol,             /* output char array (must be 16 bytes) */
   int  * portnumber );         /* output integer */

void UFDBstripURL2( 
   char * URL, 			/* input URL string */
   int    stripwwwprefix,	/* input flag for stripping "www." prefix from URL */
   char * strippedUrl,  	/* output char array (must be UFDB_MAX_URL_LENGTH bytes) */
   char * domain,       	/* output char array (must be 1024 bytes) */
   char * protocol,		/* output char array (must be 16 bytes) */
   int  * portnumber );		/* output integer */

char * UFDBprintable( char * string );

char * UFDBfgets( 
   char * requestBuffer, 
   int    bufferSize, 
   FILE * fp );

char * UFDBfgetsNoNL( char * s, int size, FILE * stream );

#endif

